import os
from stellar_plots import *


__all__ = ["driver", "mast_utils", "make_plot", "utils"]


# set Python env variable to keep track of example data dir
stellar_plots_dir = os.path.dirname(__file__) 


##################################################################################
### PURPOSE: Plot the actual output for each source
##################################################################################


##################################################################################
### IMPORT STATEMENTS
##################################################################################
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
import io
import math




##################################################################################
### PLOTTING FUNCTION
##################################################################################
def plotting(wave, flux, teff, logg, ra, dec, days, norm_flux):
    '''Plotting

    Make a figure with subplots for each of the data products pulled

    Args: 
        wave  (arr):   Numpy array. The wavelength values for the spectrum from SDSS.
        flux  (arr):   Numpy array. The flux values for the spectrum with SDSS.
        teff (float):  Float. The effective temperature from Gaia.
        lum  (float):  Float. The luminosity from Gaia. 
        ra   (float):  Float. Right Ascension of target.
        dec  (float):  Float. Declination of target.
        z    (float):  Float. Redshift of source.

    Returns:
        fig (Figure object) 
    '''


    fig = plt.figure(figsize = (10, 8))
    G = gridspec.GridSpec(20, 4)
    G.update(wspace=0.5, hspace=2.5) #include whitespace around plots

    ax1 = fig.add_subplot(G[10:14,:]) #spectrum
    ax1 = fig.axes[0]
    ax2 = fig.add_subplot(G[:8,0:2]) #HRD
    ax2 = fig.axes[1]
    ax3 = fig.add_subplot(G[:8,2:4]) #info
    ax3 = fig.axes[2]
    ax4 = fig.add_subplot(G[16:, :])
    ax4 = fig.axes[3]



    ### Spectrum
    ax1.set_xlabel(r'$\rm{Wavelength \ [\AA]}$', fontsize = 14)
    ax1.set_ylabel(r'$\mathrm{Normalized Flux}$', fontsize = 14)
    if not isinstance(wave, (np.ndarray, list, set,) ): #handle case of no available data
        ax1.set_ylim(0, 10)
        ax1.set_xlim(0, 100)
        ax1.annotate('No Spectrum Available', (25, 4), xytext = (25, 4), fontsize = 20)
    else:
        flux = [f/max(flux) for f in flux] #normalize flux for plotting
        ax1.plot(wave, flux, lw = 2, color = 'k')


    ### Light Curve
    ax4.set_xlabel(r'$\rm{Time [Days]}$', fontsize = 14)
    ax4.set_ylabel(r'$\mathrm{Normalized Flux}$', fontsize = 14)
    if not isinstance(days, (np.ndarray, list, set,) ): #handle case of no available data
        ax4.set_ylim(0, 10)
        ax4.set_xlim(0, 100)
        ax4.annotate('No Light Curve Available', (25, 4), xytext = (25, 4), fontsize = 20)
    else:
        ax4.plot(days, norm_flux, lw = 2, color = 'k')




    ### HR Diagram 
    ax2.set_xlabel(r'$\rm{T_\mathrm{eff} \ [K]}$', fontsize = 14)
    ax2.set_ylabel(r'$\mathrm{Log(g) \ [dex]}$', fontsize = 14)

    if teff < 0 or math.isnan(teff) == True or math.isnan(logg) == True: #handle case of no available data
        ax2.set_ylim(0, 10)
        ax2.set_xlim(0, 100)
        ax2.annotate('No HR Data Available', (3, 5), xytext = (3, 5), fontsize = 20)
    else:
        ax2.scatter(teff, logg, color = 'red', marker = 's', zorder = 1E6)

        #Plot the HR background of full Gaia Datasource
        full_t, full_g = np.loadtxt('./data/full_hr.txt', skiprows = 1, unpack = True)
        ax2.scatter(full_t, full_g, color = 'gray', s = 2)
        ax2.invert_xaxis()

    


    ### Info
    ax3.set_xticks([])
    ax3.set_yticks([])
    ax3.set_ylim(0, 100)
    ax3.set_xlim(0, 100)
    ax3.annotate(r"$\bf{Quick Stats: }$", (2, 90), xytext = (2, 90), fontsize = 14, color = 'k')
    ax3.annotate("Ra:  " + str(ra), (2, 80), xytext = (2, 80), fontsize = 14, color = 'k')
    ax3.annotate("Dec: " + str(dec), (2, 70), xytext = (2, 70), fontsize = 14, color = 'k')

    
    #Make a temporary file to hold the plot in
    temp = io.BytesIO()
    plt.savefig(temp, format="png")
    temp.seek(0)
    
    return temp



    


##################################################################################
### IMPORT STATEMENTS
##################################################################################
import numpy as np
import pandas as pd
import lightkurve as lk
from astropy.io import fits
from astropy import units as u
from matplotlib import pyplot as plt
from astroquery.sdss import SDSS
from astroquery.gaia import Gaia
from astroquery.vizier import Vizier
from astropy.coordinates import SkyCoord
import sys
'''
todo
- what if the input ra dec aren't in degrees
- (Isabella): handle case of wrong input file
'''

def read_table(fname):
    '''
    Read in stars from a file if it exists with columns RA & Dec.
    
    Args:
        fname (str): name of file that contains stars to read in.

    Returns:
        ra (array): right ascension (ra) values of input stars.
        dec (array): declination (dec) values of input stars.
    '''

    #ra, dec = np.loadtxt(fname, skiprows = 0, unpack = True, delimiter = ',')
    #return ra, dec
    df = pd.read_csv(fname, names=['ra','dec'], on_bad_lines='skip')
    return df['ra'].to_numpy(), df['dec'].to_numpy()
    

def load_gaia(ra, dec, dr=2):

    '''
    Read in Gaia data from Gaia DR2, and returns stellar parameters & Gaia IDs.

    Args:
        ra (float): ra values of input stars.
        dec (float): dec values of input stars.
        dr (int): Gaia data release no. Default is 2.

    Returns:
        gaia_id (int): Gaia Source IDs for all stars.
        teff (float): stellar effective temperature in Kelvin.
        rad (float): stellar radius in solar units.
        lum (float): stellar luminosity in solar units.
    '''

    if dr == 2:
        GAIA_CATALOG='I/345/gaia2'
    elif dr == 3:
        GAIA_CATALOG='I/355/gaiadr3'
    else:
        print('This GAIA version does not exist. Please specify either 2 or 3.')
        sys.exit()

    coord = SkyCoord(ra=ra, dec=dec, unit = (u.deg, u.deg))
    try:
        cat=Vizier.query_region(coord, catalog=GAIA_CATALOG, radius=2*u.arcsecond)
        cat=cat[0].to_pandas()
        return float(cat['Source'].to_numpy()), float(cat['Teff'].to_numpy()), float(cat['Rad'].to_numpy()), float(cat['Lum'].to_numpy())
    except:
        #print('Object with (RA,Dec)=(%s,%s) not found.'%(ra,dec))
        return np.nan, np.nan, np.nan, np.nan

def load_sdss(ra, dec):
    '''
    Read in Gaia data from Gaia DR2.

    Args:
        ra (float): ra of input stars.
        dec (float): dec of input stars.
    
    Returns:
        sdss_ids (array): SDSS ID of input star.
    '''

    SDSS_CATALOG='V/147/sdss12'

    coord = SkyCoord(ra=ra*u.deg, dec=dec*u.deg)
    try:
        cat=Vizier.query_region(coord, catalog=SDSS_CATALOG, radius=2*u.arcsecond)
        cat=cat[0].to_pandas().iloc[0]
        sdss_id= cat['SDSS12']
    except:
        #print('Object with (RA,Dec)=(%s,%s) not found.'%(ra,dec))
        sdss_id = np.nan
    return sdss_id

def load_kepler(ra, dec, quarters=False, cadence='long'):
    '''
    Load in photometry from Kepler if exists. Example: KIC 3733346 (287.11345099999994, 38.81283)

    Args:
        ra (float): ra values of input stars.
        dec (float): dec values of input stars.
        quarters (boolean): Kepler quarters to download. Default is False.
        cadence (str): Kepler cadence of data to use. Default is long.
    
    Returns:
        time (array): time values in lightcurve
        flux (array): flux values in lightcurve
        kicid (int): KICID of object
    '''
    coord = SkyCoord(ra=ra*u.deg, dec=dec*u.deg)
    if quarters:
        sr = lk.search_lightcurve(coord, author='Kepler', quarter=quarters, cadence=cadence)        
    else:    
        sr = lk.search_lightcurve(coord, author='Kepler', cadence=cadence)
    if len(sr) > 0:
        lcs=sr.download_all()
        lc = lc.stitch()
        lc = lc.normalize(unit='ppm').remove_nans().remove_outliers().flatten()
        kicid = int(sr[0].table['target_name'].value.data[0][4:])
        return lc.time.value, lc.flux.value, kicid
    else:
        return [np.nan],[np.nan],[np.nan]

def load_tess(ra, dec):
    '''
    Load in photometry from TESS if exists. Example: KIC 3733346 (287.11345099999994, 38.81283)
    TESS example: 261136679
    
    Args:
        ra (float): ra values of input stars.
        dec (float): dec values of input stars.

    Returns:
        time (array): time values in lightcurve
        flux (array): flux values in lightcurve
        ticid (int): TICID of object
    '''
    TICID = 261136679
    sr    = lk.search_targetpixelfile('TIC %s' % TICID)[0]
    if len(sr)>0:
        tpf = sr.download(quality_bitmask='default', author='SPOC')
        lc  = tpf.to_lightcurve()
        lc  = lc.normalize(unit='ppm').remove_nans().remove_outliers().flatten()
        return lc.time.value,lc.flux.value,tpf.targetid
    else: 
        #print('TIC %s not found.' % TICID)
        return np.nan, np.nan, np.nan
    

def get_spectrum(ra, dec):
    '''
    Returns the spectrum of input star from SDSS.

    Args:
        ra (float): ra of input stars.
        dec (float): dec of input stars.
    
    Returns:
        sp_lambda (array): wavelength in SDSS spectrum
        sp_flux (array): flux in SDSS spectrum
    '''

    coord = SkyCoord(ra=ra*u.deg, dec=dec*u.deg)

    try:
        sp = SDSS.get_spectra(coordinates = coord, radius=2*u.arcsec)[0][1]

        sp_data = sp.data

        sp_flux = sp_data.flux
        sp_lambda = 10**sp_data.loglam #lambda is returned in log space

    except(ValueError, RuntimeError, TypeError):
        return(-99, -99)

    return sp_lambda, sp_flux




##################################################################################
### PURPOSE: Main calling function
##################################################################################

##################################################################################
### PURPOSE: Import Statements
##################################################################################
from matplotlib.backends.backend_pdf import PdfPages
import sys

try:
    from make_plot import plotting as plotting
except ImportError:
    from .make_plot import plotting as plotting
try:
    from utils import *
except ImportError:
    from .utils import *
from PIL import Image
import warnings


#supress all warnings from terminal
warnings.filterwarnings("ignore")




def output_figures(input_file, output_save_dir_plus_name):

    '''Main calling function

    Main function that calls other functions to find properties for the list of sources

    Args:
        input_file (str): Points to a list (.txt) with two columns (ra, dec) and 
            1 row per source with 1 header row
        output_save_dir_plus_name(str): The path and name for where the flipbook
            should be saved on user's computer

    Returns:
        None
    '''

    pp = PdfPages(output_save_dir_plus_name) #directory to save output flipbook to
    


    #Open the coordinates from the user file
    ras, decs = read_table(input_file)

    no_sources = len(ras)

    #Loop through the coordinates for the sources
    for idx, (r, d) in enumerate(zip(ras, decs)):


        source, temp, radius, lum = load_gaia(r, d)
        sdss_id = load_sdss(r, d)
        wave, flux = get_spectrum(r, d)
        days, norm_flux, _ = load_tess(r, d)


        #Call function to make plot
        temp_img = plotting(wave, flux, temp, lum, r, d, days, norm_flux)

        #Read plot in as image and save
        fig, ax = plt.subplots(figsize = (12, 8))

        im = Image.open(temp_img)

        ax.imshow(im)
        ax.set_xticks([])
        ax.set_yticks([])

        pp.savefig(fig)

        print("Done with: " + str(idx+1) + " out of " + str(no_sources) + " Sources")
    
    pp.close()



if __name__ == "__main__":
    input = str(sys.argv[1])
    output = str(sys.argv[2])
    output_figures(input, output)




import os, sys, re, json, time
import requests
import numpy as np
import astropy.units as u
from astropy.coordinates import SkyCoord, Angle
from astroquery.vizier import Vizier
from astroquery.mast import Catalogs
from astropy.table import Table, Column, Row
from astropy.wcs import WCS

from urllib.parse import quote as urlencode
from urllib.request import urlretrieve
import http.client as httplib


def mastQuery(request):
    """Sends a request to the MAST server.
    Parameters
    ----------
    request : str
        JSON string for request.
    Returns
    ----------
    head :
        Retrieved data headers from MAST.
    content :
        Retrieved data contents from MAST.
    """
    server = 'mast.stsci.edu'

    # Grab Python Version
    version = '.'.join(map(str, sys.version_info[:3]))
        # Create Http Header Variables
    headers = {'Content-type': 'application/x-www-form-urlencoded',
               'Accept': 'text/plain',
               'User-agent': 'python-requests/'+version}
    # Encoding the request as a json string
    requestString = urlencode(json.dumps(request))
    # Opening the https cnnection
    conn = httplib.HTTPSConnection(server)
    # Making the query
    conn.request('POST', '/api/v0/invoke', 'request='+requestString, headers)

    # Getting the response
    resp = conn.getresponse()
    head = resp.getheaders()
    content = resp.read().decode('utf-8')

    # Close the https connection
    conn.close()

    return head, content

def crossmatch_by_position(pos, r, service):
    """Crossmatches [RA,Dec] position to a source in the Gaia DR2 or TIC catalog.
    Parameters
    ----------
    pos : tuple
        (RA, Dec)
    r :  float
        Radius of search for crossmatch.
    service : str
        Name of service to use. 'Mast.GaiaDR2.Crossmatch'
            or 'Mast.Tic.Crossmatch' are accepted.
    Returns
    -------
    table : astropy.table.Table
        Table of crossmatch results.
    """

    crossmatchInput = {'fields': [{'name':'ra' , 'type':'float'},
                                  {'name':'dec', 'type':'float'}],
                       'data': [{'ra':pos[0], 'dec':pos[1]}]}
    request = {'service':service,
               'data':crossmatchInput,
               'params': {'raColumn':'ra', 'decColumn':'dec', 'radius':r},
               'format':'json', 'removecache':True}
    headers, outString = mastQuery(request)
    return jsonTable(json.loads(outString))

def crossmatch_distance(pos, match):
    """Returns distance in arcsec between two sets of coordinates."""
    c1 = SkyCoord(pos[0]*u.deg, pos[1]*u.deg, frame='icrs')
    c2 = SkyCoord(match[0]*u.deg, match[1]*u.deg, frame='icrs')
    return c1.separation(c2).to(u.arcsec)

def jsonTable(jsonObj):
    """Converts JSON return type object into an astropy Table.
    Parameters
    ----------
    jsonObj :
        Output data from `mastQuery`.
    Returns
    ----------
    dataTable : astropy.table.Table
    """
    dataTable = Table()
    for col,atype in [(x['name'],x['type']) for x in jsonObj['fields']]:
        if atype=='string':
            atype='str'
        if atype=='boolean':
            atype='bool'
        if atype=='int':
            atype='float'
        dataTable[col] = np.array([x.get(col,None) for x in jsonObj['data']],dtype=atype)
    return dataTable

def tic_from_coords(coords):
    """Returns TIC ID, Tmag, and separation of best match(es) to input coords."""
    tess = crossmatch_by_position(coords, 0.01, 'Mast.Tic.Crossmatch')
    tessPos = [tess['MatchRa'], tess['MatchDEC']]
    sepTess = crossmatch_distance(coords, tessPos)
    subTess = tess[sepTess==np.min(sepTess)]
    return int(subTess['MatchID'].data[0]), [subTess['Tmag'].data[0]], sepTess[sepTess==np.min(sepTess)]/u.arcsec, int(subTess['version'][0]), subTess['contratio'].data[0]


def gaia_from_coords(coords):
    """Returns Gaia ID of best match(es) to input coords."""
    gaia = crossmatch_by_position(coords, 0.01, 'Mast.GaiaDR2.Crossmatch')
    gaiaPos = [gaia['MatchRA'], gaia['MatchDEC']]
    sepGaia = crossmatch_distance(coords, gaiaPos)
    return int(gaia[sepGaia==np.min(sepGaia)]['MatchID'].data[0])